@@ -110,7 +110,9 @@ def self_test_suite():
         'hooks',
         'ignore',
         'index',
+        'line_ending',
         'lru_cache',
+        'mailmap',
         'objects',
         'objectspec',
         'object_store',
@@ -123,6 +125,7 @@ def self_test_suite():
         'refs',
         'repository',
         'server',
+        'stash',
         'utils',
         'walk',
         'web',
@@ -133,6 +136,13 @@ def self_test_suite():
 
 
 def tutorial_test_suite():
+    import dulwich.client  # noqa: F401
+    import dulwich.config  # noqa: F401
+    import dulwich.index  # noqa: F401
+    import dulwich.reflog  # noqa: F401
+    import dulwich.repo  # noqa: F401
+    import dulwich.server  # noqa: F401
+    import dulwich.patch  # noqa: F401
     tutorial = [
         'introduction',
         'file-format',
@@ -145,22 +155,23 @@ def tutorial_test_suite():
 
     def setup(test):
         test.__old_cwd = os.getcwd()
-        test.__dulwich_tempdir = tempfile.mkdtemp()
-        os.chdir(test.__dulwich_tempdir)
+        test.tempdir = tempfile.mkdtemp()
+        test.globs.update({'tempdir': test.tempdir})
+        os.chdir(test.tempdir)
 
     def teardown(test):
         os.chdir(test.__old_cwd)
-        shutil.rmtree(test.__dulwich_tempdir)
+        shutil.rmtree(test.tempdir)
     return doctest.DocFileSuite(
+            module_relative=True, package='dulwich.tests',
             setUp=setup, tearDown=teardown, *tutorial_files)
 
 
 def nocompat_test_suite():
     result = unittest.TestSuite()
     result.addTests(self_test_suite())
+    result.addTests(tutorial_test_suite())
     from dulwich.contrib import test_suite as contrib_test_suite
-    if sys.version_info[0] == 2:
-        result.addTests(tutorial_test_suite())
     result.addTests(contrib_test_suite())
     return result
 
@@ -175,7 +186,7 @@ def compat_test_suite():
 def test_suite():
     result = unittest.TestSuite()
     result.addTests(self_test_suite())
-    if sys.version_info[0] == 2 and sys.platform != 'win32':
+    if sys.platform != 'win32':
         result.addTests(tutorial_test_suite())
     from dulwich.tests.compat import test_suite as compat_test_suite
     result.addTests(compat_test_suite())
